      program mm5decode

c ---------------------------------------------------------------------
c --- MM5DECODE -- Unpack and extract MM5 3D packed output tiles
c ---              and create a 3D.DAT file (in Version 2.1 format)
c
c                  This is the official program that extracts the
c                  packed MM5 tiles for Europe, South America - Chile/
c                  Argentina region, South America - Northern region
c                  South America - Brazil
c
c --- MM5DECODE    Version: 3.4       Level: 100118                  MAIN
c
c     Copyright (c) 2010 by Exponent, Inc.
c
c --- Zhong-Xiang Wu
C     Developed: 12/28/2007

C From Version: 3.3  Level: 091013 to Version: 3.4  Level: 100118 (Minor)
C  1. Revise the way to fix the turelat1 for PST map projection 
C     in more general way insteat of just fixing Greenland tiles 

C From Version: 3.2  Level: 090930 to Version: 3.3  Level: 091013 (Minor)
C  1. Add available year section in TILES.CFG file.
C  2. When full tile name is used, adjust the key name for different 
C     years
C
C From Version: 3.1  Level: 090613 to Version: 3.2  Level: 090930 (Major)
C  1. Add flag for tile file naming format, so that both simple and 
C     full file names can be accepted.
C  2. TILES.CFG file format changed.  Add standard file header lines,
C     keys of full tile names, and brief descriptions of MM5 domains
C
C  Zhong Wu
c  9/30/2009

C From Version: 3.0  Level: 090402 to Version: 3.1  Level: 090613 (Minor)
C  1. Fix an bug for turelat1 for PST map projection for Greenland tiles,
C     Use "PS" instead of "PST" in cmap
C
C  Zhong Wu
c  6/13/2009

C From Version: 2.6  Level: 090309 to Version: 3.0  Level: 090402 (Major)
C  1. Use configuration file to replace all header files
C  2. Use long tile names
C  3. Initialize ix=0 to avoid possible Nan
C  4. Add n3dd variable to accout possible missing water contents 
C     in vertical profiles.  Change dataver3d='2.1' to dataver3d='2.2'
C     to account possible five zeros or -5 compression in missing water 
C     contents case
C  5. Revise wrtcmp subroutine. Set wind direction to 360 degree 
C     instead of 0 degree for wind speed above zero 
C  6. Remove un-used subroutine
C  7. Add special header treatment fix Greenland PS projection header
C     info
C
C Zhong-Xiang Wu
C 4/2/2009

C From Version: 2.5  Level: 080923 to Version: 2.6  Level: 090309 (Major)
C  1. Reduce SSTBASE from 250K to 200K due to low SST (<250K) in 
C     EPA 12-km MM5 Output. Corresponding changes were made in
C     RUCDECODE (V2.6)
C  2. Change PACK.DAT Data Version to dataver = '2.2'
C
C Zhong-Xiang Wu
C 3/9/2009

C From Version: 2.4  Level: 080910 to Version: 2.5  Level: 080923 (Major)
C  1. Allow number of tiles 1000 and more
C  2. Add Thailand, Australia_East/West, New Zealand, US, Canada, 
C     and Saudi Arabia 2001.  Brazil was already coded as Region 6.
C
C Zhong-Xiang Wu
C 9/23/2008

C From Version: 2.3  Level: 080825 to Version: 2.4  Level: 080910 (Minor)
C  1 Add three 12-km domains: Peru, Middle East, and China-East
C
C Zhong-Xiang Wu
C 9/10/2008

C From Version: 2.2  Level: 080730 to Version: 2.3  Level: 080825 (Minor)
C  1 Add domain 2 (12-km) MM5 data sets for Japan, South Africa
C
C Zhong-Xiang Wu
C 7/30/2008

C From Version: 2.2  Level: 080730 to Version: 2.3  Level: 080825 (Minor)
C  1 Add domain 2 (12-km) MM5 data sets for Japan, South Africa
C
C Zhong-Xiang Wu
C 7/30/2008

C From Version: 2.1  Level: 080701 to Version: 2.2  Level: 080730 (Major)
C  1 Change tile setup for later MM5 packing in following way:
C    i. 10 x 10 grids for internal tiles
C   ii. For last row and column:  The maximum size for 10 x 10 grids tiles
C       in 495.2 MB from MM5EU-2006 tiles. The size limit of one tile
C       is 750 MB.  Therefore the maximum grids holding by one tile
C       is 150 grids
C       a. Equal to, or less than, 50 grids, merge to next inner tiles
C       b. More than 50 grids, separate tiles themselves
C  2. This code still accept previous tile configuration for the four 
C     previous MM5 runs (EUD2-2006/2007, SAMS_D2, SAMN_D2) 
C  3. Add Europe (Italy) Domain 3 (4-km) and Domain 4 (1.333-km)
C 
C Zhong Wu
C 7/30/2008

C From Version: 2.0  Level: 080522 to Version: 2.1  Level: 080701 (Minor)
C     1. Add South America-Brizal
C     2. Keep original header info
C
C Zhong Wu
C 7/1/2008

C From Version: 1.2  Level: 080512 to Version: 2.0  Level: 080522 (Major)
C     1. Decode for all three MM5 regions (Europe, Chile/Argentina, 
C        South America-North)
C     2. Add comline call
C
C Zhong Wu
C 5/22/2008

C From Version: 1.1  Level: 080416 to Version: 1.2  Level: 080512 (Minor)
C     1. Settings for Northern South America MM5 domain2 
C
C Zhong Wu
C 5/12/2008

C From V1.0 Level 071228 to Version: 1.1  Level: 080416 (Minor)
C     1. Add parameters of nxdmn and nydmn
C     2. Change way of settings for mxnx,mxny,ni2,nj2
C     3. Settings for SAMS mm5 domain 2 (3D.DAT header file and 
C        lat/lon ranges). 
C
C Zhong Wu
C 4/16/2008


C --- Output files are in the format of 3D.DAT Version 2.1
      include 'mm5decode.par'
      include 'mm5decode.cm1'

c----------------------------------------------------------------------
c
c --- Set Code Version and Level number of code
      cver='3.4'
      clevel='100118'

c --- Set Input Packed Dataset Version and Level number for this code
      dataset1='3DPACK.DAT'
      dataver1='2.2'

      read(dataver1,*)fver1
c      print *,' fver1:',fver1

c --- Set 3D.DAT Dataset Version and Level number fit for this code
      dataset3d='3D.DAT'
      dataver3d='2.2'

c --- Open control file with user input information
      fct='mm5decode.inp'
      call comline(fct)

      open(ict,file=fct,status='old')
c
c --- Read title line from control file
      read(ict,101)title_inp
 101  format(a)
c
c --- Read output list file name 
      read(ict,101)flog
      nt=index(flog,' ')-1
      flog=flog(1:nt)
c
c --- Open output list file
      open(ilog,file=flog,status='unknown')
c
c --- Write version and level number of code to list file and
c --- echo back control file inputs
      write(ilog,301)cver,clevel
301   format(//1x,'MM5DECODE -- Version: ',a16,' Level: ',a16//)

      write(ilog,351)dataset1,dataver1,fver1
351   format(//1x,'Input dataset: ',a16,' Dataver: ',a16,2x,f10.2,//)

      write(ilog,302)title_inp
302   format(1x,'Title: ',a80)
      write(ilog,303)flog
303   format(/1x,'List file name: ',a80)

      read(ict,*)id_reg
      if(id_reg.lt.1) then
         write(ilog,*)' Error: MM5 Regional ID outrange: ',id_reg
         print *,' Error: MM5 Regional ID outrange: ',id_reg
         stop
      endif

      write(ilog,361)id_reg
 361  format(' MM5 Selected Region ID: ',i2)

c --- Read drive & path where tiled data is stored 
c --- (trailing backslash (\) needed on PC)
      read(ict,101)cdev
      ntdev=index(cdev,' ')-1
      cdev=cdev(1:ntdev)
      if(ntdev.eq.0) then
         ntdev=1
         cdev=' '
      endif
      write(ilog,304)cdev(1:ntdev)
304   format(/1x,'Input packed data from drive/path: ',a)
c
c --- Read output 3D.DAT file name
      read(ict,101)fou
      ntfou=index(fou,' ')-1
      fou=fou(1:ntfou)
      write(ilog,305)fou(1:ntfou)
305   format(/1x,'Output 3D.DAT file: ',a)

c --- Read/write method of selecting domain (1= I/J; 2=long/lat)
      read(ict,*)iselect
      write(ilog,306)iselect
306   format(/1x,'Method for selecting domain: ',i6,
     1 ' (1=I/J; 2=lon/lat)')

      if(iselect.eq.1) then
         read(ict,*)ibeg,iend
         read(ict,*)jbeg,jend
         write(ilog,*)ibeg,iend,jbeg,jend
      else if(iselect.eq.2)then
         read(ict,*)flonbeg,flonend
         read(ict,*)flatbeg,flatend
         write(ilog,*)flonbeg,flonend,flatbeg,flatend
      else
         write(ilog,314)iselect
314      format(/1x,'Invalid selection method for defining domain: ',i4/
     1   1x,'Selection method must be: 1 (I/J) or 2 (long/lat)')
         print *,errmes
         stop
      endif

      read(ict,*)kbeg,kend

      write(ilog,*)kbeg,kend

c --- Read beginning and ending dates/times (UTC) to extract
c     (Format: yyyymmddhh)
      read(ict,*)idatebeg
      read(ict,*)idateend

      write(ilog,*)idatebeg,idateend

c --- Read/write batch/interactive mode switch 
c    (1=batch processing, 0=interactive)
      read(ict,*)ibatch
      write(ilog,322)ibatch
322   format(/1x,'Batch or interactive mode processing switch: ',
     1     i2,' (1=batch mode, 0 = interactive mode)')

c --- QA checks on batch/interactive mode switch
      if(ibatch.ne.0.and.ibatch.ne.1)then
         write(ilog,323)ibatch
323      format(/1x,'ERROR -- Invalid value for batch/interactive ',
     1   'flag -- IBATCH = ',i6/1x,'Valid values are 0 or 1')
         print *,errmes
         stop
      endif

      read(ict,*)iway
      write(ilog,333)iway
 333  format(/1x,'Tile naming format (1=Simple, 2 = Full): ', i2)

      if(iway.lt.1 .or. iway.gt.2) then
         write(ilog,*)' Illegal tile file format:',iway
         write(*,*)' Illegal tile file format:',iway
         stop
      endif

c --- Reading control file finished
      close(ict)

c --- Echo beginning and ending dates/times (UTC) to extract
c     (Format: yyyymmddhh)
      write(ilog,319)idatebeg
319   format(/1x,'Beginning date/time to extract (UTC): ',i10,
     1 ' (yyyymmddhh)')
      write(ilog,320)idateend
320   format(1x,'   Ending date/time to extract (UTC): ',i10,
     1 ' (yyyymmddhh)')

c --- QA checks on dates/time to extract
      if(idatebeg.gt.idateend) then
         write(ilog,321)idatebeg,idateend
321      format(/1x,'ERROR -- Beginning date/time is after ending ',
     1   'date/time -- ','IDATEBEG: ',i10,3x,'IDATEEND: ',i10)
         print *,errmes
         stop
      endif

C --- Get number of hours
      call gethours(ilog,idatebeg,idateend,nhrtot)
      write(ilog,340)nhrtot
 340  format(/1x,'Total number of hours of data to extract: ',i6)

C --- Determine tiled data year beased on beg/ending time
      call getdate(idatebeg,iyb,imb,idb,ihb)
      call getdate(idateend,iye,ime,ide,ihe)
      nyrpk=(iyb+iye)/2

C --- Special treatment for some short periods
      if(nyrpk.eq.2005 .and. ime.eq.12 .and. ihe.le.23) nyrpk=2006
      if(nyrpk.eq.2000 .and. ime.eq.12 .and. ihe.le.23) nyrpk=2001

      write(ilog,*)' Year for tiled data:',nyrpk

C --- Configuration info
      call configmm5

      if(iselect.eq.1) goto 2000

C --- Convert lat/lon input to X/Y for iselet=2
      call convert
      call setij

 2000 continue

C --- QA check for control inputs
      if(iselect.eq.1) then
         write(ilog,371)
 371     format(/1x,'Using I/J to select a region',
     2   1x,'(Negative longitude for Western Hemisphere)/')
         write(ilog,*)' Required I/J Ranges:',ibeg,iend,jbeg,jend
         if(ibeg.gt.iend) then
            write(ilog,308)ibeg,iend
 308        format(/1x,'ERROR -- Beginning I > ending I -- IBEG: ',i6,
     1      1x,'IEND: ',i6)
            print *,errmes
            stop
         endif
         if(jbeg.gt.jend) then
            write(ilog,309)jbeg,jend
 309        format(/1x,'ERROR -- Beginning J > ending J -- JBEG: ',i6,
     1      1x,'JEND: ',i6)
            print *,errmes
            stop
         endif

         if(ibeg.lt.nx1 .or. iend.gt.nx2) then
            write(ilog,307)nx1,nx2,ibeg,iend
            write(*,307)nx1,nx2,ibeg,iend
 307        format(1x,'Error: I - Outrange: MM5 ranges: ',2i4,
     &           /,1x,'           Reguired: ',2i4)
            stop
         endif
         if(jbeg.lt.ny1 .or. jend.gt.ny2) then
            write(ilog,3071)ny1,ny2,jbeg,jend
            write(*,3071)ny1,ny2,jbeg,jend
 3071       format(1x,'Error: J - Outrange: MM5 ranges: ',2i4,
     &           /,1x,'           Reguired: ',2i4)
            stop
         endif
      else if(iselect.eq.2)then
         write(ilog,311)
 311     format(/1x,'Using Beginning/Ending longtiudes/latitudes '
     1   'to select a region'/
     2   1x,'(Negative longitude for Western Hemisphere)/')

         write(ilog,312)flonbeg,flonend
 312     format(1x,'Seleted longitude range: ',f8.3,2x,f8.3)
         write(ilog,313)flatbeg,flatend
 313     format(1x,'Seleted latitude range: ',f8.3,2x,f8.3)
      else
         write(ilog,324)iselect
 324     format(/1x,'Invalid selection method for defining domain: ',i4/
     1   1x,'Selection method must be: 1 (I/J) or 2 (long/lat)')
         print *,errmes
         stop
      endif

c --- Vertical range of layers
      write(ilog,315)kbeg,kend
315   format(/1x,'Seleted range of layers to extract: ',i6,3x,i6)
c
c --- QA checks on layer selections
      if(kbeg.gt.kend) then
         write(ilog,316)kbeg,kend
316      format(/1x,'ERROR -- Beginning layer > ending layer -- ',
     1   'KBEG: ',i6,1x,'KEND: ',i6)
         print *,errmes
         stop
      endif
c
c --- Warning if first layer is not Layer #1
      if(kbeg.ne.1)then
         write(ilog,317)kbeg
317      format(/1x,'Error -- First layer is NOT Layer #1',
     1   'KBEG = ',i6)
         print *,'Error -- First layer is NOT Layer #1 -- KBEG = ',
     1   kbeg
         print *,'Processing stopped'
         stop
      endif
c
c --- Error if top layer is > MXZ
      if(kend.gt.nz2)then
         write(ilog,318)kend,nz2
318      format(/1x,'ERROR -- Top layer is > MXZ -- KEND = ',i6,3x,
     1   'MXZ = ',i6)
         print *,errmes
         stop
      endif

C --- Determine tiles needed
      call gettiles

c --- Get log and scratch file needed info -----
      ntsch=index(fou,'.',back=.true.)-1
      if(ntsch.le.0) ntsch=ntfou

c
c --- Open output 3D.DAT file
      open(io,file=fou,status='unknown')
      
c --- Open scratch files -----
      idatebkp=idatebeg

      do i=1,ncd
         isc=isch+i
         open(isc,status='scratch',form='unformatted')
      enddo

c --- Process every CDs needed ----------------------
c --- First create scratch files for each CD
      id_3dpack=1
      id_chkbeg=1
      n3dd=n3d

      do kcd=1,ncd
         iout=isch+kcd

c        I/J range on this CD
         icd=ncdpk(kcd)

         nx1=ijcdpk(1,kcd)   ! I/J ranges on this CD 
         nx2=ijcdpk(2,kcd)
         ny1=ijcdpk(3,kcd)
         ny2=ijcdpk(4,kcd)

         i1=ijcdpk(5,kcd)     ! Selected I/J ranges on this CD
         i2=ijcdpk(6,kcd)
         j1=ijcdpk(7,kcd)
         j2=ijcdpk(8,kcd)

         write(ilog,*)' I/J ranges: ',i1,i2,j1,j2
         print *,' I/J ranges: ',i1,i2,j1,j2

         if(ibatch.eq.0) then
            print *
            print *,' Is the CD ready? (Y/y)'
            write(ilog,*)' Is the CD ready? (Y/y)'
            read(*,'(a)')CDY
            write(ilog,'(a)')cdy
         else
            CDY='y'
         endif

         if(cdy.ne.'Y' .and. cdy.ne.'y') then
            write(ilog,*)' CD is not ready. Process aborting'
            print *,' CD is not ready. Process aborting'
            stop
         else
            write(ilog,330)kcd,icd
330         format(/1x,'Processing File: ',i4.4,' Tile #: ',i4.4)
            print 2073,kcd,icd
2073        format(' Processing File: ',i4.4,' Tile #: '
     &       ,i4.4,' Please Wait......')
         endif

c        Open input CD file
         if(ndig.le.3) then
            if(iway.eq.1) then
               write(fin,2060)cdev(1:ntdev),icd
            else
               write(fin,2070)cdev(1:ntdev),trim(tilekey),icd
            endif
 2060       format(a,'Tile',i3.3,'.pck')
 2070       format(2a,'_Tile',i3.3,'.pck')
         elseif(ndig.le.4) then
            if(iway.eq.1) then
               write(fin,2160)cdev(1:ntdev),icd
            else
               write(fin,2170)cdev(1:ntdev),trim(tilekey),icd
            endif
 2160       format(a,'Tile',i4.4,'.pck')
 2170       format(2a,'_Tile',i4.4,'.pck')
         else
            print *,' Error: Tiles Outrange (1-9999) - ',ndig
            stop
         endif

         write(ilog,*)' Input File Name ',trim(fin)
         print *, ' Input File Name ',trim(fin)

         open(in,file=fin,status='old',form='binary',action='read')  !unix
c         open(in,file=fin,status='old',form='unformatted'
c     &       ,access='transparent',action='read')  !pc

C ---    Read first line to check packed tile format    
C ---    Read header info on packed tile first
         if(kcd.eq.1) then
            idout=1
         else
            idout=0
         endif

         if(id_3dpack.eq.1) call rdhd(idout,id_3dpack)

         write(ilog,*)' id_3dpack:',id_3dpack

         if(kcd.eq.ncd .and. id_3dpack.eq.1) call outgeo

         nx1=ijcdpk(1,kcd)   ! I/J ranges on this CD 
         nx2=ijcdpk(2,kcd)
         ny1=ijcdpk(3,kcd)
         ny2=ijcdpk(4,kcd)

         i1=ijcdpk(5,kcd)     ! Selected I/J ranges on this CD
         i2=ijcdpk(6,kcd)
         j1=ijcdpk(7,kcd)
         j2=ijcdpk(8,kcd)

         ihr=0
         iprt=0
         idatenext=idatebeg

c         print *,icd,nx1,nx2,ny1,ny2
 5000    continue
         ifirst=1
         iprt=iprt+1
         itlbeg=1
         idate_b=-9999

         do j=ny1,ny2
            do i=nx1,nx2

               read(in,err=6000)idate,(izips(m),m=1,5)

C ----         QA for date stamp
               if(i.eq.nx1 .and. j.eq.ny1) then
                  idate_b=idate
c                  print *,' idate=',idate
                  if(id_chkbeg.eq.1 .and. idate.gt.idatebeg) then
                     print *,' Error: Reguested date too early'
                     print *,'        Requested date:',idatebeg
                     print *,'        Beg. date in File:',idate
                     stop
                  else
                     id_chkbeg=0
                  endif
               else
                  if(idate.ne.idate_b) then
                     write(ilog,4021)idate_b,idate,i,j
                     write(*,4021)idate_b,idate,i,j
4021                 format('Error: Date/Time Mismatch Between Grids:'
     &                 ,/,' Time at 1st grid in tile:',i12,/,
     &                 ' Time at Grid I/J:',i12,2i4)
                     stop 106
                  endif
               endif

C	       Make sure no-negative values
               id_stop=1
	       do m=1,5
                  izip=izips(m)
                  call chknneg(izip,ilog,id_stop)
               enddo

               if(idate.ne.idatenext .and.
     &            i.eq.nx1 .and. j.eq.ny1) then
                  write(ilog,261)icd,idate,idatenext
c                  print 261,icd,idate,idatenext
 261              format(' Skip  : Tile/Date/Datenext - ',i4.4,2i12)
c                  stop
               elseif(idate.eq.idatenext .and.
     &            i.eq.nx1 .and. j.eq.ny1) then
c                  write(ilog,262)icd,idate,idatenext
c                  print 262,icd,idate,idatenext
 262              format(' Select: Tile/Date/Datenext - ',i4.4,2i12)
               endif

               if(i.ge.i1 .and. i.le.i2 .and.
     &            j.ge.j1 .and. j.le.j2 
     &            .and. idate.eq.idatenext) then
                  lpk=1

C                 Figure out next date
                  if(ifirst.eq.1) then
                     ifirst=0
                     ihr=ihr+1
                     call getdate(idate,iy,im,id,ih)
                     call chgtim(iy,im,id,ih,1)
                     call timestmp(iy,im,id,ih,idatenext_new)
c                     print *,iprt,ihr,idate,idatenext_new
                  endif
               else
                  lpk=0
               endif

               if(lpk.eq.1) then
                  write(iout)idate,izips
               endif             

               do k=1,nz
                  read(in,err=6000)idata
C                 QA check data
                  id_stop=0
                  do m=1,3
                     call chknneg(idata(m),ilog,id_stop)               
                  enddo
                  if(lpk.eq.1) write(iout)idata
               enddo
  
               read(in,err=6000)np,(ix2(ii),ii=1,np)

               if(lpk.eq.1) write(iout)np,(ix2(ii),ii=1,np)

c               if(i.eq.nx2 .and. j.eq.ny2) then
c                  print *,' idate2=',idate
c               endif

            enddo
         enddo

         if(ifirst.eq.0) idatenext=idatenext_new
         idatebkp=idate
c
c ---    Write a message every 1000 hours
         if(mod(iprt,1000).eq.0) then
            write(ilog,2061)icd,iprt,idate
            print 2061,icd,iprt,idate
 2061       format(9x,'Tile: ',i4.4, ' at Hr ',i5,2x,i12)
         endif

         if(ihr.lt.nhrtot) goto 5000

         write(ilog,*)' Last hour processed:',idate
         print *,' Last hour processed:',idate
         goto 6500

C ---    Read Error processing 
 6000    print *,' Not Enough Data in Input Files'
         print *,' Last Date/Time Can Be Processed:',idate
         write(ilog,*)' Not Enough Data in Input Files'
         write(ilog,*)' Last Date Can Be Processed:',idate

         idateend=idatebkp

         call gethours(ilog,idatebeg,idateend,nhrtot)
         print 334,nhrtot
         write(ilog,334)nhrtot
334      format(/1x,' Total number of hours Changed to: ',i6)

 6500    continue

      enddo   ! end of kcd (finish reading all CDs and create scratch files)

C --- Rewind scratch files
      print *,'Creating 3D.DAT now, Please Wait......'
      do i=1,ncd
         isc=isch+i
         rewind(isc)
      enddo

C --- Decode, order, and merge all scratch files
      do 8000 ihr=1,nhrtot

         do jrep=1,nyrep
            ny1=ijcdpk(7,jrep)  ! use J range in first one
            ny2=ijcdpk(8,jrep)

            if(jrep.eq.1) then
               ny10=ny1
               ny20=ny2
            else
               if(ny1.ne.ny20+1) then
                  print *,' Error in NY1 and NY02: ',ny1,ny20
                  stop
               endif
            endif

            do j=ny1,ny2

               do irep=1,nxrep            ! Change I range 
                  kcd=(irep-1)*nyrep+jrep
                  inp=isch+kcd

                  nx1=ijcdpk(5,kcd) ! Selected I/J ranges on this CD
                  nx2=ijcdpk(6,kcd) ! Assume rectanglar, equal size in vertical

                  if(irep.eq.1) then
                     nx10=nx1
                     nx20=nx2
                  else
                     if(nx1.ne.nx20+1) then
                        write(ilog,*)'Error in NX1 and NX02: ',nx1,nx20
                        print *,' Error in NX1 and NX02: ',nx1,nx20
                        stop
                     endif
                  endif

                  do i=nx1,nx2
                     read(inp)idate,(izips(m),m=1,5)
c                     print *,' chk:',idate,(izips(m),m=1,5)

                     call decode_datahd(idate,i,j,ii,jj)
                     call chkhd(idate,ii,jj,ierr)

                     write(io,1070)idate,ii,jj,psldot,rain,isnow,
     &                    radsw,radlw,t2,qq2,wd10,ws10,sst
 1070                format(i10.10,2i3,f7.1,f5.2,i2,3f8.1,f8.2,3f8.1)

c Decimal: P-0, z-0,t-1,wd-0,ws-1,w-2,rh-0,q/cld/rn/ic/sn/gr-2
c Unit: P-mb, z-m,t-K,wd-o,ws-m/s,w-m/s,rh-%,q/cld/rn/ic/sn/gr-g/kg
c              unpack and output variables p,z,t,wd,ws,w,rh,q
 
                     id_stop=1
                     do k=1,nz
                        read(inp)idata

c                       Decoding word 1 (HHHHHPPPTT)               
c                       idata(1)=izz*ifive+ipp*100+itt2
                        izip=idata(1) ! first word
                        call chknneg(izip,ilog,id_stop)

                        izz=izip/ifive
                        ipp=izip/100-izz*1000
                        itt2=izip-izz*ifive-ipp*100

C                       Account for negative height (>= -500 m)   
                        if(id_3dpack.eq.1) then
                           if(k.le.40) then
                              hbase=hbase1
                           else
                              hbase=hbase2
                           endif
                        else
                           if(k.le.40) then
                              hbase=0
                           else
                              hbase=hbase2
                           endif
                        endif

                	izz=izz+nint(hbase)	! unit: m

C ----                  Pressure decode and QA check
                        if(k.le.30) then
                          kp=ipp+pbase1
                        else
                          kp=ipp+pbase2
                        endif

c                       Decoding word 2 (TTWDWWSWS)
c                       idata(2)=itt1*ieight+iwd*ifive+iws*10+iww1
                        izip=idata(2)
                        call chknneg(izip,ilog,id_stop)

                        itt1=izip/ieight
                        iwd=izip/ifive-itt1*1000
                        iws=izip/10-itt1*iseven-iwd*ifour
                        iww1=izip-itt1*ieight-iwd*ifive-iws*10
                        ws=iws/10.
                        itt=itt1*100+itt2
                        tk=itt/10.+tkbase

c                       Decoding word 3 (RRHQQQQWWW)
c                       idata(3)=irh*iseven+iqq*ithree+iww2
                        izip=idata(3)
                        call chknneg(izip,ilog,id_stop)

                        irh=izip/iseven
                        iqq=izip/ithree-irh*ifour
                        iww2=izip-irh*iseven-iqq*ithree
                        qq=iqq/100.
                        iww=iww1*1000+iww2
                        ww=iww/100.
                        if(ww.ge.wbase) ww=wbase-ww

c                       Put in array 
                        kpp(k)=kp
                        kh(k)=izz
                        ftk(k)=tk
                        fws(k)=ws
                        kwd(k)=iwd
                        krh(k)=irh
                        fqq(k)=qq
                        fww(k)=ww

                     enddo

                     call chkorder(idate,ii,jj,ierr)

c                    Unpack and output variables cld,rn,ic,sn,gr
c                    Note: packed ix2 starting from surface
                     read(inp)np,(ix2(ii),ii=1,np)

                     ipos=1
                     ip=0
                     ivar=1
                     ik=1

                     iclmt=nz*5
                     if(iclmt.lt.1) then
                        print *,' Illegal iclmt:',iclmt
                        print *,' iclmt must be >= 1 '
                        stop
                     endif

                     ix=0
 1000                ic=ix2(ipos)

C                    Account some values > 32767 in ix2
                     if(ic.lt.-iclmt) then
                        ictmp1=ic
                        ic=ic+65536
                        ictmp2=ic
C                       Set to the cap value same as used 
C                           in rucencode (PACKv2) 
                        if(ic.gt.32000) ic=32000  

                        write(ilog,510)idate,i,j
                        write(ilog,511)ictmp1,ictmp2,ic
 510                    format('Warning: Water/ice con. out range at'
     &                          ,3i12)
 511                    format('Org: ', i8,', Actual:',i8,', Set to',i8)
                     endif

                     if(ic.ge.0) then
                        ip=ip+1
                        ivar=(ip-1)/nz+1 ! vertical first
                        ik=ip-(ivar-1)*nz
c                       ik=(ip-1)/nvar+1  ! variable first
c                       ivar=ip-(ik-1)*nvar

c                        iks=nz-ik+1
c                        ix(ivar,iks)=ic
                        ix(ivar,ik)=ic
c                       print *,' >0 ',nvar,ip,ik,ivar,ic
                     else
                        nn=-ic
                        ipos=ipos+1
                        ic=ix2(ipos)

                        do ii=1,nn
                           ip=ip+1
                           ivar=(ip-1)/nz+1 ! vertical first
                           ik=ip-(ivar-1)*nz
c                          ik=(ip-1)/nvar+1 ! Variable first
c                          ivar=ip-(ik-1)*nvar

c                           iks=nz-ik+1
c                           ix(ivar,iks)=ic
                           ix(ivar,ik)=ic
c                          print *,' <0 ',nvar,ip,ik,ivar,ic
                        enddo
                     endif

                     ipos=ipos+1

                     if(ipos .le. np) goto 1000

c ----               QA range check for 3D profile
                     if(ip.ne.nz*5) then
                        nerr=nerr+1
                        write(ilog,404)ip,nz*5
                        print 404,ip,nz*5
404                     format(1x,'Error: Vertical levels Mismatch: '
     &                            ,2i5)
                        stop 301
                     endif

                     do k=1,nz
                        xlvl(1)=kpp(k)
                        xlvl(2)=kh(k) 
                        xlvl(3)=ftk(k)
                        xlvl(4)=kwd(k)
                        xlvl(5)=fws(k)
                        xlvl(6)=fww(k)
                        xlvl(7)=krh(k)
                        xlvl(8)=fqq(k)

                        xlvl(9)=ix(1,k)/1000.
                        xlvl(10)=ix(2,k)/1000.
                        xlvl(11)=ix(3,k)/1000.
                        xlvl(12)=ix(4,k)/1000.
                        xlvl(13)=ix(5,k)/1000.
            
                        call chklvl(idate,ii,jj,k,ierr)
                     enddo

c ---                Write out in MM5 format
                     do k=kbeg,kend

C...                 Compress the last five fields
                     xcmp=0

                     xcmp(1)=kpp(k)
                     xcmp(2)=kh(k) 
                     xcmp(3)=ftk(k)
                     xcmp(4)=kwd(k)
                     xcmp(5)=fws(k)
                     xcmp(6)=fww(k)
                     xcmp(7)=krh(k)
                     xcmp(8)=fqq(k)

                     do ii=9,n3dd
                        jj=ii-8
                        xcmp(ii)=ix(jj,k)/1000.
                     enddo

                     call wrtcmp(xcmp,n3dd)

                     enddo  ! end of k

                  enddo   ! end nx1-nx2

                  nx10=nx1
                  nx20=nx2

               enddo    ! end 1-nxrep
            enddo     ! end ny1-ny2

            ny10=ny1
            ny20=ny2

         enddo     ! end 1-nyrep

 8000 continue   ! end 1-nhrtot   

C --- Close scratch files
      do i=1,ncd
         isc=isch+i
         close(isc)
      enddo

      print *,'----      Extraction succeeded        ----'
      write(ilog,*)
      write(ilog,*)'----      Extraction succeeded        ----'

      stop
      end

      include 'mm5decode.blk'
      include 'calutils.for'
      include 'coordlib.for'

c ---------------------------------------------------------------------
      subroutine getdate(ndate,iyr,imon,iday,ihour)
c
c --- MM5DECODE    Version: 3.4       Level: 050507             GETDATE

      iyr=ndate/1000000
      imon=ndate/10000-iyr*100
      iday=ndate/100-iyr*10000-imon*100
      ihour=ndate-iyr*1000000-imon*10000-iday*100
      
      return
      end

c ----------------------------------------------------------------------
      subroutine wrtcmp(xcmp,n3dd)

c --- MM5DECODE    Version: 3.4       Level: 090403               WRTCMP
c                  Zhong-Xiang Wu

C     Purpose:
C     Write one level 3D profile in compressed form
C     (Compressed output of MM5.dat data record for ioutform >= 93)
C
C     Zhong-Xiang Wu
C     4/2/2009

      parameter (xzero=0.0049)

      include 'mm5decode.par'

      dimension xcmp(n3d)

      ipre=nint(xcmp(1))
      izh=nint(xcmp(2))
      atk=xcmp(3)
      iwd=nint(xcmp(4))
      aws=xcmp(5)
      awz=xcmp(6)
      irh=nint(xcmp(7))
      aq=xcmp(8)

      if(iwd.eq.0 .and. aws.gt.0.01) iwd=360 

C --- Compress
      idcmp=1
      ncmp=0
      do i=9,n3dd
         if(xcmp(i).ge.xzero) then
            idcmp=0
            goto 1000
         else
            ncmp=ncmp+1
         endif
      enddo

 1000 continue

      if(idcmp.eq.1) then
         write(io,95)ipre,izh,atk,iwd,aws,awz,irh,aq,float(-ncmp)
      else
         write(io,95)ipre,izh,atk,iwd,aws,awz,irh,aq,
     &       (xcmp(i),i=9,n3dd)
      endif
 95   format( i4,i6,f6.1,i4,f5.1,f6.2,i3,f5.2,5f6.3)

      return
      end

! **********************************************************************
      subroutine chgtim(iyr,imon,iday,ihour,idt)

c --- MM5DECODE    Version: 3.4       Level: 060823               
! --- Zhong-Xiang Wu           
!
! --- PURPOSE:
!     Increase or decrease hours. 
!     Used to convert Local Stand Time to GMT or to setup time stamp for loop

      parameter(nmonth=12)

      dimension ndays(nmonth)

      data ndays/31,28,31,30,31,30,31,31,30,31,30,31/

      if(mod(iyr,4).eq.0) then
         ndays(2)=29 
      else
         ndays(2)=28
      endif

      ihour=ihour+idt

      if(idt.lt.0) goto 2000

 1000 if(ihour.gt.23) then
         ihour=ihour-24
         iday=iday+1
         
         if(iday.gt.ndays(imon)) then
            iday=1
            imon=imon+1
            
            if(imon.gt.12) then
               imon=1
               iyr=iyr+1
               if(iyr/4 .eq. iyr/4.0) then
                  ndays(2)=29 
               else
                  ndays(2)=28
               endif
            endif
         endif
         goto 1000
      else
         return
      endif

 2000 continue

 3000 if(ihour.lt.0) then
         ihour=ihour+24
         iday=iday-1
         
         if(iday.le.0) then
            imon=imon-1
            if(imon.le.0) then
               iyr=iyr-1
               imon=12
               if(iyr/4 .eq. iyr/4.0) then
                  ndays(2)=29 
               else
                  ndays(2)=28
               endif
            endif
   
            iday=ndays(imon)
         endif
         goto 3000
      else
         return
      endif

      end
! --------------------------------------------------
      Subroutine timestmp(iyr,imn,idy,ihr,idate)

c --- MM5DECODE    Version: 3.4       Level: 060823               

      idate=ihr+idy*100+imn*10000+iyr*1000000

      return
      end

c --------------------------------------------------------------
	subroutine gethours(ilog,ndate1,ndate2,nhrs)

c --- MM5DECODE    Version: 3.4       Level: 061228               
C Purpose:
c       Get hours between two dates
C Zhong Wu
C (3/22/1999)

C Changes
C  1. Fix an error for total hours when beginning year is 
C     leap year
C  Zhong Wu      
C  5/23/2003

c	Get hours between to dates

	if(ndate2.lt.ndate1) then
	   print *,'Error in gethours: ndate2 < ndate1'
	   print *,ndate1,ndate2
	   stop
	endif

	call getdate(ndate1,iyr1,imon1,iday1,ihour1)
	call julday(ilog,iyr1,imon1,iday1,jday1)

	call getdate(ndate2,iyr2,imon2,iday2,ihour2)
	call julday(ilog,iyr2,imon2,iday2,jday2)

        call deltt(iyr1,jday1,ihour1,iyr2,jday2,ihour2,nhrs)

        nhrs=nhrs+1   ! deltt is one hour less total hours
 
	return
	end

C ------------------------------------------------------------------
      Subroutine rdhd(idout,id_3dpack)

c --- MM5DECODE    Version: 3.4       Level: 090309               

C Purpose:
C    Read header info on each packed tile
C
C Zhong-Xiang Wu
C 12/28/2007

C Update:
C     1. Add n3dd variable
C 
C Zhong Wu
C 4/3/2009

      include 'mm5decode.par'
      include 'mm5decode.cm1'

C --- Check the packed data format
      read(in)dataset,dataver,datamod

      read(dataver,*)fver
c      print *,' fver=',fver

      if(dataset.ne.dataset1 .or. fver.gt.fver1) then
c      if(dataset.ne.dataset1 .or. fver.lt.fver1) then
         write(ilog,2011)
 2011    format(/,' Error: No Header on tiles is not allowed ')
         stop
      else
         write(ilog,2110)
 2110    format(/,' 3DPACK Format with Header Info',/)
         print 2110
      endif

      if(fver.lt.2.2) then
         sstbase=sstbase1
      else
         sstbase=sstbase2
      endif

      write(ilog,2111)sstbase,dataver,fver
 2111 format('SSTBASE:',f10.2,2x,a,2x,f10.2)

      datamod2='Created from Packed 3DPACK.DAT '//trim(dataver)
      if(idout.eq.1) write(io,1010)dataset3d,dataver3d,datamod2
1010  format(2a16,a64)

C...  Read Header Record #2
C     Note: New format header is considered as comment lines, not real
C           record. Keep it there for later use
      read(in)ncomm
      if(idout.eq.1) then
         ncomm1=ncomm+1
         write(io,1011)ncomm1
1011     format(i4)

         write(commenthd,1110)cver,clevel
1110     format('Produced by MM5DECODE Version: ',a8,
     &          ', Level: ',a8)
         write(io,1111)commenthd
1111     format(a80)
      endif

C...  Read Header Record #3  ! read/write original comment lines
      do i=1,ncomm
         read(in)commenthd

c         if(i.eq.ncomm-1) write(commenthd,1311)
c 1311    format('3D.DAT was created from 12-km MM5')
c         if(i.eq.ncomm) commenthd='Original '//trim(commenthd)

         if(idout.eq.1) write(io,1111)commenthd
      enddo

C --- New part of header (As comment lines in 3D.DAT file) 
C --- Beginning of new part
      read(in)model1,model2,model3
c      if(idout.eq.1) write(io,1212)model1,model2,model3      
1212  format(4a16)

      read(in)pmap
      read(in)clat0,clon0,clat1,clat2

C --- Special treatment for Greenland PS header info
      if(id_reg.ge.33 .and. id_reg.le.35) then
         pmap='PS'
         clat1='70.0000N'
      endif

c      if(idout.eq.1) write(io,1113)pmap
1113  format(a8)
c      if(idout.eq.1) write(io,1212)clat0,clon0,clat1,clat2

      read(in)feast,fnorth
c      if(idout.eq.1) write(io,1115)feast,fnorth
1115  format(2f9.3)

      read(in)datum
c      if(idout.eq.1) write(io,1113)datum

C     NIMA date
      read(in)daten
c      if(idout.eq.1) write(io,1116)daten
1116  format(a12)

C     Horizontal map unit: XUNIT
      read(in)xunit
c      if(idout.eq.1) write(io,1117)xunit
1117  format(a4)

C     Time zone (UTM+0000)
      read(in)tzone
c      if(idout.eq.1) write(io,1118)tzone
1118  format(a8)

C --- MM5 parameters-domain settings
      read(in)x1dmn,y1dmn,dxy,nx,ny,nz
c      if(idout.eq.1) write(io,1213)x1dmn,y1dmn,dxy,nx,ny,nz
 1213 format(2f10.3,f8.3,2i4,i3)

C --- End of new part of header line as comments

C...  Read Header Record #4
      read(in)ioutw,ioutq,ioutc,iouti,ioutg,iosrf
      if(idout.eq.1) write(io,1012)ioutw,ioutq,ioutc
     &    ,iouti,ioutg,iosrf
 1012 format (6i3)

      if(ioutg.eq.0) n3dd=n3d-1
      if(ioutg.eq.0 .and. iouti.eq.0) n3dd=n3d-3

C...  Read Header Record #5
      cmap=trim(pmap)
      call rdlat(clat0,rlatc)
      call rdlon(clon0,rlonc)

      call rdlat(clat1,truelat1)
      call rdlat(clat2,truelat2)

C --- Fix an truelat1 bug in calmm5 version 2.12 and before for PST proj. 
      if(cmap.eq.'PST ') then
         truelat1=rlatc
         cmap="PS  "
      endif

      if(idout.eq.1) write(io,1013)cmap,rlatc,rlonc,truelat1
     &        ,truelat2,x1dmn,y1dmn,dxy,nx,ny,nz
 1013	format(a4,f9.4,f10.4,2f7.2,2f10.3,f8.3,2i4,i3)

C...  Read Header Record #6
      read(in)inhyd,imphys,icupa,ibltyp,ifrad,isoil,
     &         ifddaan,ifddaob,igrdt,ipbl,ishf,ilhf,iustr,iswdn,
     &         ilwdn,ist1,ist2,ist3,ist4,ist5,ist6

      if(idout.eq.1) write(io,1014)inhyd,imphys,icupa,ibltyp,ifrad,
     &   isoil,ifddaan,ifddaob,igrdt,ipbl,ishf,ilhf,iustr,iswdn,
     &   ilwdn,ist1,ist2,ist3,ist4,ist5,ist6
 1014 format (30(i3))

C --- Hold output of Record #8, sigma level, and lat/lon later 
C...  Read Header Record #7 (subdomain only)
      read(in)idatetmp,nhours,nxsub,nysub,nz

C...  Read Header Record #8
      read(in)nx1,ny1,nx2,ny2,nz1,nz2,rxmin,rxmax,rymin,rymax

C...  Next nz records:sigma levels 
      read(in)(hsig(k),k=1,nz)

C...  Geophysical records. 
C...  Note: lat/lon and elev at dot point, land use at cross point
      do j=ny1,ny2
         do i=nx1,nx2
            read(in)ii,jj,rlat,rlon,ihgt,iland,rlatcrs
     &           ,rloncrs,ihgtcrs

            flatdot(ii,jj)=rlat
            flondot(ii,jj)=rlon
            flatcrs(ii,jj)=rlatcrs
            floncrs(ii,jj)=rloncrs
            lev(ii,jj)=ihgt
            land(ii,jj)=iland
            levcrs(ii,jj)=ihgtcrs
            if(ii.ne.i .or. jj.ne.j) then
               print *,' I/J match Error: ',ii,i,jj,j
               write(ilog,326)ii,i,jj,j
326            format(' Error: Header I/J not match (II,I,JJ,J)- ',4i5)
               stop
            endif

C ---       Check data
c            write(19,191)ii,jj,flatdot(ii,jj),flondot(ii,jj),lev(ii,jj)
c     &                   ,land(ii,jj)
c191         format(2i5,2f11.4,2i6)
         enddo
      enddo

3000  continue

C --- Check consistancy of region and tiles
      if(rlatc.ne.rlatch .or. rlonc.ne.rlonch .or.
     &   truelat1.ne.truelat1h .or. truelat2.ne.truelat2h) then
         write(ilog,336)rlatc,rlonc,truelat1,truelat2,
     &                  rlatch,rlonch,truelat1h,truelat2h
         write(*,336)rlatc,rlonc,truelat1,truelat2,
     &                  rlatch,rlonch,truelat1h,truelat2h
 336     format(' Error: Tile/Header Mapping mis-Match - ',/,
     &          '        Tile Mapping:   ',4f12.5,/, 
     &          '        Header Mapping: ',4f12.5) 
         stop
      endif

      return
      end

C ---------------------------------------------------------------
      Subroutine outgeo

c --- MM5DECODE    Version: 3.4       Level: 061228               

C --- Purpose:
C     Output geo part in 3D.DAT header
C
C Zhong-Xiang Wu
C 11/28/2008

      include 'mm5decode.par'
      include 'mm5decode.cm1'

C --- Make sure lat/lon info from tiles is large enough
      if(ibeg.lt.imi .or. iend.gt.imx .or.
     &   jbeg.lt.jmi .or. jend.gt.jmx) then
         write(ilog,101)ibeg,iend,jbeg,jend,imi,imx,jmi,jmx
         print 101,ibeg,iend,jbeg,jend,imi,imx,jmi,jmx
101      format(' Error: Domain not match:',8i5)
         stop
      endif

      if(kbeg.lt.nz1 .or. kend.gt.nz2) then
         write(ilog,102)kbeg,kend,nz1,nz2
         print 102,kbeg,kend,nz1,nz2
102      format(' Error: Sigma level not match:',8i5)
      endif

C     Min/max lat/lon for selected domain
      rxmin1=99999
      rxmax1=-99999
      rymin1=99999
      rymax1=-99999

      do j=jbeg,jend
         do i=ibeg,iend
            rlat=flatdot(i,j)
            rlon=flondot(i,j)
            if(rlat.lt.rymin1) rymin1=rlat
            if(rlat.gt.rymax1) rymax1=rlat
            if(rlon.lt.rxmin1) rxmin1=rlon
            if(rlon.gt.rxmax1) rxmax1=rlon
         enddo
      enddo
    
      nxp=iend-ibeg+1
      nyp=jend-jbeg+1
      nzp=kend-kbeg+1

C --- Record #7
      write(io,1015)idatebeg,nhrtot,nxp,nyp,nzp
 1015 format(i10,i5,3i4)

C --- Record #8
      write(io,1016)ibeg,jbeg,iend,jend,kbeg,kend,rxmin1
     &      ,rxmax1,rymin1,rymax1
 1016 format(6i4,2f10.4,2f9.4)

C     Sigma level
      write(io,1022)(hsig(k),k=kbeg,kend)
 1022 format(f6.3)

C     Lat/lon
      do j=jbeg,jend
         do i=ibeg,iend
            rlat=flatdot(i,j)
            rlon=flondot(i,j)
            rlatcrs=flatcrs(i,j)
            rloncrs=floncrs(i,j)
            ihgt=lev(i,j)
            iland=land(i,j)
            ihgtcrs=levcrs(i,j)

            write(io,1023)i,j,rlat,rlon,ihgt,iland,rlatcrs
     &           ,rloncrs,ihgtcrs
1023        format(2i4,f9.4,f10.4,i5,i3,1x,f9.4,f10.4,i5)

         enddo
      enddo

      return
      end


C -----------------------------------------------------
	Subroutine rdlat(clat,rlat)

c --- MM5DECODE    Version: 3.4       Level: 061228               

C Purpose:
C    Write character latitude
C Zhong Wu
C 12/20/2006
     
      real rlat
      character*16 clat

      nt=len_trim(clat)      
      read(clat(1:nt-1),*)rlat
      if(clat(nt:nt).eq.'S' .or. clat(nt:nt).eq.'s') rlat=-rlat

      return
      end

C -----------------------------------------------------
	Subroutine rdlon(clon,rlon)

c --- MM5DECODE    Version: 3.4       Level: 061228               

C Purpose:
C    Write character longitude
C Zhong Wu
C 12/20/2006
     
      real rlon
      character*16 clon

      nt=len_trim(clon)      
      read(clon(1:nt-1),*)rlon
      if(clon(nt:nt).eq.'W' .or. clon(nt:nt).eq.'w') rlon=-rlon

      return
      end
C ------------------------------------------------------------------
         Subroutine chknneg(izip,ilog,id_stop)

c --- MM5DECODE    Version: 3.4       Level: 061228               

C Purpose:
C    Check no-negative integer
C Zhong-Xiang Wu
C 12/08/2006

	integer*4 izip

        if(izip.lt.0) then
           write(ilog,101)izip
101        format('Error: Integer*4 < 0', i12)
           if(id_stop.eq.1) stop 129
        endif

	return
	end
C ------------------------------------------------------------------
	Subroutine decode_datahd(idate,i,j,ii,jj)

c --- MM5DECODE    Version: 3.4       Level: 061228               

C Purpose:
C    Decode data header recode
C Zhong-Xiang Wu
C 12/08/2006

        include 'mm5decode.par'
        include 'mm5decode.cm1'
    
        integer*4 izip1,izip2,izip3,izip4,izip5

C ---   Check I/J consistency at each grid
	izip1=izips(1)
	izip2=izips(2)
	izip3=izips(3)
	izip4=izips(4)
	izip5=izips(5)

c        print *,izip1,izip2,izip3,izip4,izip5
C	Word1: I/J (JJJIII)
        ii=izip1/1000
        jj=izip1-ii*1000
        if(ii.ne.i .or. jj.ne.j) then
           write(ilog,403)idate,ii,jj,i,j
           write(*,403)idate,ii,jj,i,j
 403       format('Error: I/J Mismatch at ',i12,/,
     &            'I/J in file: ',2i4, 'I/J should be:',2i4)
           stop 130
        endif

C ----  Word 2: SLP, Rain, ISNOW(PPPPPRRRRS)
        ipsl=izip2/ifive
        irain=izip2/10-ipsl*ifour
        isnow=izip2-ipsl*ifive-irain*10

        psldot=ipsl/10.
        rain=irain/100.

C ----  Word 3: SW, LW (SSSSSLLLLL)
        iradsw = izip3/ifive
        iradlw2 = izip3 - iradsw*ifive   ! avoid round-up err

        radsw = iradsw/10.
        radlw2 = iradlw2/10.
        radlw  = radlw2 - 1000.

C ----  Word 4: T2, QQ2 (S1TTTTQQQQ)
        nsst1 = izip4/ieight
        it2    = (izip4-nsst1*ieight)/ifour
        iqq2   = izip4-nsst1*ieight-it2*ifour
        t2 = it2/10.
        qq2 = iqq2/100.

C ----  Word 5: WSWSWDWDST (SSSSDDDDST)
        iws102 = izip5/isix
        iwd10  = izip5/100 - iws102*ifour
        nsst2 = izip5 - iws102*isix - iwd10*100

        ws102=iws102/10.
        ws10 = ws102 - 50.
        wd10=iwd10/10.

c        print *,idate
c        print *,izip5,iws102,iwd10,ws102,ws10,wd10

        nsst = nsst2 + nsst1*100
        sst = nsst/10. + sstbase
c        print *,nsst1,nsst2,nsst,sstbase,sst

        xhd(1)=psldot
        xhd(2)=rain
        xhd(3)=isnow

C ---   Set MM5 SW/LW to zero to be consistent with RUC when MM5 data
C       were used to replace RUC missing periods
        xhd(4)=radsw
        xhd(5)=radlw

        xhd(6)=t2
        xhd(7)=qq2
        xhd(8)=wd10
        xhd(9)=ws10
        xhd(10)=sst

	return
	end
C ----------------------------------------------------------------------

      Subroutine chkhd(idate,ii,jj,ierr)

c --- MM5DECODE    Version: 3.4       Level: 061228               

C Purpose:
C    QA for data header record in 3D.DAT file
C
C Zhong-Xiang Wu
C 10/26/2006

      include 'mm5decode.par'
      include 'mm5decode.cm1'
    
      character*6 cname

      ierr=0 
      do ivar=1,n2d
        alow=alim2d(1,ivar)
        ahig=alim2d(2,ivar)

        aa=xhd(ivar)
        if(aa.lt.alow .or. aa.gt.ahig) then
           ierr=ierr+1
           cname=cname2d(ivar)
           write(ilog,101)cname,idate,ii,jj,aa,alow,ahig
           write(*,101)cname,idate,ii,jj,aa,alow,ahig
101        format(1x,'Warning: Outrange(2D) - ',a6, i12,2i4,3f14.4)
        endif
      enddo

      return
      end

C ----------------------------------------------------------------------

      subroutine chkorder(idate,ii,jj,ierr)

c --- MM5DECODE    Version: 3.4       Level: 061228               

C Purpose:
C     1. Check vertical order of P and GH
C     2. Check pressure difference of Level 1 and 2
C
C Zhong-Xiang Wu
C 10/26/2006

      include 'mm5decode.par'
      include 'mm5decode.cm1'

      ierr=0
      ip1=kpp(nz1)
      ih1=kh(nz1)
      hter=lev(ii,jj)

      if(ih1.lt.0) then
           write(ilog,124)cname3d(2),idate,ii,jj,ih1
124        format(1x,'Warning: GH < 0 at level1: ',a6,i12,2i4,i6,
     &     /,' But fixed later')
      endif

C --- Pressure disorder between first/send levels in RUC data
      do k=nz1+1,nz2
        ip2=kpp(k)
        ih2=kh(k)

        if(k.eq.nz1+1) then
           dpp=ip1-ip2
        else
           dpp=0
        endif

C       Note: ip1=ip2 is possible in RUC at high level
        if(ip2.gt.ip1 .or. dpp.gt.dppc) then
           write(ilog,101)cname3d(1),idate,ii,jj,k,ip1,ip2,dpp
           write(*,101)cname3d(1),idate,ii,jj,k,ip1,ip2,dpp
101        format(1x,'Error: P-Vertical order:',a6,i12,3i4,2i6,f7.1)
           ierr=ierr+1

           stop 203
        endif

        if(ih2.lt.ih1) then
           write(ilog,102)cname3d(1),idate,ii,jj,k,ih1,ih2
           write(*,102)cname3d(1),idate,ii,jj,k,ih1,ih2
           ierr=ierr+1
           stop 204
        endif
102     format(1x,'Error: GH-Vertical order:',a6,i12,3i4,2i6)

        if(ih1.le.hter) then
           write(ilog,103)cname3d(2),idate,ii,jj,k-1,ih1,hter
           write(*,103)cname3d(2),idate,ii,jj,k-1,ih1,hter
           ierr=ierr+1
           stop 205
        endif
103     format(1x,'Error: GH <= Terrain:',a6,i12,3i4,i6,f9.3)
        
        ip1=ip2
        ih1=ih2

      enddo

      return
      end

C ---------------------------------------------------------------
      Subroutine chklvl(idate,ii,jj,kk,ierr)

c --- MM5DECODE    Version: 3.4       Level: 090403               

C Purpose:
C    QA for one vertical level in 3D.DAT file
C
C Zhong-Xiang Wu
C 10/26/2006

      include 'mm5decode.par'
      include 'mm5decode.cm1'
    
      character*6 cname

c      do ivar=1,n3dd
c         print *,' in call:',ivar,xlvl(ivar)
c      enddo

      ierr=0
      do ivar=1,n3dd
        alow=alim3d(1,ivar)
        ahig=alim3d(2,ivar)
        aa=xlvl(ivar)

        if(aa.lt.alow .or. aa.gt.ahig) then
           ierr=ierr+1
           cname=cname3d(ivar)
           write(ilog,101)cname,idate,ii,jj,kk,aa,alow,ahig
           write(*,101)cname,idate,ii,jj,kk,aa,alow,ahig
101        format(1x,'Warning: Outrange(3D) - ',a6, i12,3i4,3f14.4)
        endif
      enddo

      return
      end

C --------------------------------------------------------------------
      Subroutine configmm5

c --- MM5DECODE    Version: 3.4      Level: 100118               

C Purpose:
C    Get configuration information
C
C Zhong-Xiang Wu
C 4/2/2009

C From Version: 3.3  Level: 091013 to Version: 3.4  Level: 100118 (Minor)
C  1. Revise the way to fix the turelat1 for PST map projection 
C     in more general way insteat of just fixing Greenland tiles 

      parameter(iecho=1)

      include 'mm5decode.par'
      include 'mm5decode.cm1'

      character*300 buff1,buff2,buff3

C --- Open three configuration files
      open(ifg,file=fcfg,status='old',action='read')  ! MM5 config

C --- QA check for config files
      call lastup(ifg,cvercfg,clvlcfg,lastupdate)

C --- Skip header comment line
      read(ifg,*)ncom
      do i=1,ncom
         read(ifg,*)
      enddo

      read(ifg,101)header1
 101  format(a)

C --- Read tile configuration part
 1000 continue
      read(ifg,101)buff1
      read(buff1,*)id

C --- Find the region needed
C --- Domain Info
      if(id.eq.id_reg) then

         read(buff1,301)id,cmap,rlatch,rlonch,truelat1h,truelat2h,
     &        x1dmn,y1dmn,dxy,nx,ny,nz,nx1,ny1,nx2,ny2,nz1,nz2,
     &        rxmin,rxmax,rymin,rymax
 301     format(i3.3,1x,a4,f9.4,f10.4,2f7.2,2f10.3,f8.3,2i4,i3,6i4
     &       ,2f10.4,2f9.4)
         read(buff1(9:),*)(cfg(i),i=1,nas)

C ---     Fix an truelat1 bug in calmm5 version 2.12 and before for PST proj.
C         for Greenland runs
         if(cmap.eq.'PST ') then
            cfg(3)=rlatch
            truelat1h=rlatch
            cmap="PS  "
         endif

         write(ilog,'(a)')trim(header1)
         write(ilog,301)id,cmap,rlatch,rlonch,truelat1h,truelat2h,
     &        x1dmn,y1dmn,dxy,nx,ny,nz,nx1,ny1,nx2,ny2,nz1,nz2,
     &        rxmin,rxmax,rymin,rymax
      endif

      if(id.ne.idend) goto 1000

C --- Read Tile name and description part
      read(ifg,*)  ! skip on sub_title line

 2000 continue
      read(ifg,101)buff1
      read(buff1,*)id

C --- Find the region needed
C --- Domain Info
      if(id.eq.id_reg) then
         read(buff1(5:),'(a60)')tilekey
         read(buff1(65:),'(a)')tiledes

C        Modify tilekey for different year
         write(tilekey(5:8),'(i4.4)')nyrpk

         write(ilog,311)trim(tilekey)
         print 311,trim(tilekey)
         write(ilog,312)trim(tiledes)
         print 312,trim(tiledes)
 311     format(' Tile Naming:       ',a)
 312     format(' Model Description: ',a)
         goto 3000
      endif

      if(id.ne.idend) goto 2000

 3000 continue

C --- Tile Configuration for entire domain
      ni2=nx2
      nj2=ny2

      nxd=ni2-ni1+1
      nyd=nj2-nj1+1

      ncx=nxd/igrd   ! Number of CD in I dir.
      ncy=nyd/jgrd   ! Number of CD in J dir.

C --- Earlier packing for Europe_D2 (2006/2007), Chile/Argentina, Northern SA 
C     use old tile configuration (may have narrow tiles in last column/row)
      if(id_reg.le.3) then
         if(mod(nxd,igrd).ne.0) ncx=nxd/igrd+1
         if(mod(nyd,jgrd).ne.0) ncy=nyd/jgrd+1      
      else
C ---    Later packing used new tile configuration (no narrow tiles 
C        in last column/row)
         modx=mod(nxd,igrd)
         mody=mod(nyd,jgrd)
         
         modx=modx*jgrd
         mody=mody*igrd

         if(modx.gt.modc) ncx=ncx+1
         if(mody.gt.modc) ncy=ncy+1
      endif

      ncdtot=ncx*ncy
      if(ncdtot.lt.1000) then
         ndig=3
      elseif(ncdtot.lt.10000) then
         ndig=4
      else
         print *,' Error: Tiles exceed 10000'
         stop
      endif

      write(ilog,201)ni1,ni2,nj1,nj2,ncx,ncy,ncdtot
      write(*,201)ni1,ni2,nj1,nj2,ncx,ncy,ncdtot
 201  format('ni1,ni2,nj1,nj2,ncx,ncy,ncdtot=',8i6)

      if(nx1.gt.ni1 .or. ny1.gt.nj1) then
         write(ilog,202)nx1,ny1,ni1,nj1
         write(*,101)nx1,ny1,ni1,nj1
 202     format(' Error: nx1/ny1 > ni1,nj1 - ',4i5)
         stop
      endif

      close(ifg)

      return
      end

C -------------------------------------------------------------------
      Subroutine convert

c --- MM5DECODE    Version: 3.4      Level: 090402               

C --- Purpose:
C     Convert lon/lat to X/Y 
C 
C --- Zhong-Xiang Wu
C     1/12/2009

      include 'mm5decode.par'
      include 'mm5decode.cm1'
      
c --- Initialize GLOBE1
      cmapi='LL'
      iutmi=999.9
      tmsone=1.0
      rlat1i=cfg(3)  
      rlat2i=cfg(4)
      rlat0i=cfg(1)
      rlon0i=cfg(2)
      feasti = 0.
      fnorti = 0.
      
      cmapo=cmap
      iutmo=999.9
      tmsone=1.0
      rlat1=cfg(3)
      rlat2=cfg(4)
      rlat0=cfg(1)
      rlon0=cfg(2)
      feasto = 0.
      fnorto = 0.

c --  Set conversion vectors
      call GLOBE1(
     &  cmapi,iutmi,tmsone,rlat1i,rlat2i,rlat0i,rlon0i,feasti,fnorti,
     &  cmapo,iutmo,tmsone,rlat1,rlat2,rlat0,rlon0,feasto,fnorto,
     &  caction,vecti,vecto)

c --- Convert station coordinates to model map projection and datm,
c     Translate coordinates 
      datumi='WGS-84  '
      datumo='NWS-84  '

      call GLOBE(ilog,caction,datumi,vecti,datumo,vecto,
     &     flonbeg,flatbeg,xbeg,ybeg,iutmo,c4hem)
      call GLOBE(ilog,caction,datumi,vecti,datumo,vecto,
     &     flonend,flatend,xend,yend,iutmo,c4hem)

c      write(ilog,122)flonbeg,flonend,flatbeg,flatend,xbeg,xend,ybeg,yend
c 122  format('Lon/Lat & X/Y beg-end:',8f12.4)
      write(ilog,*)'Extraction range converted from Lon/Lat to X/Y ---'
      write(ilog,*)' Beg/End Lat (deg): ',flatbeg, flatend
      write(ilog,*)' Beg/End Lon (deg): ',flonbeg, flonend
      write(ilog,*)' Beg/End X (km): ',xbeg,xend
      write(ilog,*)' Beg/End Y (km): ',ybeg,yend

      return
      end

C --------------------------------------------------------
      Subroutine setij

c --- MM5DECODE    Version: 3.4      Level: 090402               

C Purpose:
C    Get I/J range for lat/lon selection
C
C Zhong-Xiang Wu
C 4/2/2009

      include 'mm5decode.par'
      include 'mm5decode.cm1'

      x1dmn=cfg(5)  ! coarse SW grid Xo
      y1dmn=cfg(6)  ! coarse SW grid Yo
      dxy=cfg(7)      ! fine grid size used as new 3D grid size

      fi1=(xbeg-x1dmn)/dxy+1
      fi2=(xend-x1dmn)/dxy+1
      fj1=(ybeg-y1dmn)/dxy+1
      fj2=(yend-y1dmn)/dxy+1

      i1=int(fi1)
      i2=int(fi2)
      j1=int(fj1)
      j2=int(fj2)

      if(float(i2).lt.fi2) i2=i2+1
      if(float(j2).lt.fj2) j2=j2+1

      ibeg=i1
      iend=i2
      jbeg=j1
      jend=j2

      write(ilog,330)ibeg,iend,jbeg,jend
      print 330,ibeg,iend,jbeg,jend
 330  format(' I/J ranges converted from Lon/Lat:',4i6)

C --- Adjust I/J beg/end
      if(ibeg.lt.nx1) then
         write(ilog,331)ibeg,nx1
         print 331,ibeg,nx1
 331     format(' Warning: Select I1 < NX1, adjusted to NX1:',2i5)
         ibeg=nx1
      endif
      if(iend.gt.nx2) then
         write(ilog,332)iend,nx2
         print 332,iend,nx2
 332     format(' Warning: Select I2 > NX2, adjusted to NX2:',2i5)
         iend=nx2
      endif

      if(jbeg.lt.ny1) then
         write(ilog,333)jbeg,ny1
         print 333,jbeg,ny1
 333     format(' Warning: Select J1 < NY1, adjusted to NY1:',2i5)
         jbeg=ny1
      endif

      if(jend.gt.ny2) then
         write(ilog,334)jend,ny2
         print 334,jend,ny2
 334     format(' Warning: Select J2 > NY2, adjusted to NY2:',2i5)
         jend=ny2
      endif

      nn1=ibeg*iend
      nn2=jbeg*jend
      if(nn1.lt.1 .or. nn2.lt.1) then
         write(ilog,336)nn1,nn2,ibeg,iend,jbeg,jend
         print 336,nn1,nn2,ibeg,iend,jbeg,jend
 336     format(' Error: At least one negative for I/J range:',/,6i10)
         stop
      endif

      nni=iend-ibeg+1
      nnj=jend-jbeg+1
      if(nni.lt.1 .or. nnj.lt.1) then
         write(ilog,335)nni,nnj,ibeg,iend,jbeg,jend
         print 335,nni,nnj,ibeg,iend,jbeg,jend
 335     format(' Error: No grid Selected:',6i6)
         stop
      endif

      return
      end

C ----------------------------------------------------------------
      Subroutine gettiles

c --- MM5DECODE    Version: 3.4      Level: 090402               
C
C --- Purpose:
C     Determine which tiles are needed
C
C --- Zhong-Xiang Wu
C     4/2/2009

      include 'mm5decode.par'
      include 'mm5decode.cm1'
      
C --- Determine which tiles are needed
c      write(ilog,*)'I/J Ranges on each tile'

C     Figure out I/Js for this tile
C     West to East, then South to North

      do i=1,ncx     ! I/J for all tiles
         i1=ni1+(i-1)*igrd
         i2=i1+igrd-1
         if(i.eq.ncx) i2=ni2

         do j=1,ncy
            icd=(i-1)*ncy+j

            j1=nj1+(j-1)*jgrd
            j2=j1+jgrd-1
            if(j.eq.ncy) j2=nj2

            ijcd(1,i,j)=i1
            ijcd(2,i,j)=i2
            ijcd(3,i,j)=j1
            ijcd(4,i,j)=j2


c            if(ndig.le.3) then
c               write(ilog,202)icd,i,j,i1,i2,j1,j2
c               write(*,202)icd,i,j,i1,i2,j1,j2
c            else
c               write(ilog,203)icd,i,j,i1,i2,j1,j2
c               write(*,203)icd,i,j,i1,i2,j1,j2
c            endif
 202        format(' Tile-',i3.3,10i6)
 203        format(' Tile-',i4.4,10i6)

        enddo
      enddo

      print *,' ibeg/iend:',ibeg,iend
      print *,' jbeg/jend:',jbeg,jend

      do i=1,ncx  ! columns needed
         i1=ijcd(1,i,1)
         i2=ijcd(2,i,1)

         if(ibeg.ge.i1 .and. ibeg.le.i2) icd_i1=i
         if(iend.ge.i1 .and. iend.le.i2) icd_i2=i
      enddo

      do j=1,ncy  ! rows needed
         j1=ijcd(3,1,j)
         j2=ijcd(4,1,j)

         if(jbeg.ge.j1 .and. jbeg.le.j2) icd_j1=j
         if(jend.ge.j1 .and. jend.le.j2) icd_j2=j
      enddo

      write(ilog,327)
327   format(/1x,'Tiles needed for selected domain'/
     1 1x,' No.',2x,'Tile #',5x,'I/J Ranges on Tile',5x,
     2 'I/J Ranges Selected')
 
      nyrep=icd_j2-icd_j1+1
      nxrep=icd_i2-icd_i1+1

      ncd=0
      imi=999
      imx=-999
      jmi=999
      jmx=-999

      do i=icd_i1,icd_i2
        do j=icd_j1,icd_j2               
            icd=(i-1)*ncy+j
            ncd=ncd+1
            ncdpk(ncd)=icd  ! CD order #

c           I/J ranges needed on these CDs
            i1=ijcd(1,i,j)
            i2=ijcd(2,i,j)
            j1=ijcd(3,i,j)
            j2=ijcd(4,i,j)

            ijcdpk(1,ncd)=i1   ! I beg on this CD
            ijcdpk(2,ncd)=i2   ! I end on this CD
            ijcdpk(3,ncd)=j1   ! J beg on this CD
            ijcdpk(4,ncd)=j2   ! J end on this CD

            if(i1.le.imi) imi=i1
            if(i2.ge.imx) imx=i2
            if(j1.le.jmi) jmi=j1
            if(j2.ge.jmx) jmx=j2

            if(i1.lt.ibeg) i1=ibeg
            if(i2.gt.iend) i2=iend
            if(j1.lt.jbeg) j1=jbeg
            if(j2.gt.jend) j2=jend

            ijcdpk(5,ncd)=i1   ! picked I beg on this CD
            ijcdpk(6,ncd)=i2   ! picked I end on this CD
            ijcdpk(7,ncd)=j1   ! picked J beg on this CD
            ijcdpk(8,ncd)=j2   ! picked J end on this CD

            write(*,328)ncd,ncdpk(ncd),(ijcdpk(ii,ncd),ii=1,8)
            write(ilog,328)ncd,ncdpk(ncd),(ijcdpk(ii,ncd),ii=1,8)
328         format(1x,i4,4x,i4.4,4x,4i5,3x,4i5)
         enddo
      enddo

      write(ilog,329)ncd
      print 329,ncd
329   format(/1x,'Total Number of tiles needed: ',i6)

      write(ilog,339)imi,imx,jmi,jmx
339   format(/1x,'I/J ranges on all needed CDs: ',4i4)

      return
      end
C --------------------------------------------------------------------
      Subroutine lastup(in,cvercfg,clvlcfg,lastupdate)

c --- MM5DECODE    Version: V3.2       Level: 090930                 
c
c --- Purpose:
C     Get lastupdating date timestamp
c 
c --- Zhong-Xiang Wu
C     9/29/2009

      character*120 buff

      read(in,'(a)')buff

      read(buff,'(2a16)')cvercfg,clvlcfg
      nt=index(buff,':')+1
      read(buff(nt:),*)lastupdate
      
      return
      end
